"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.makeValidateAsync = exports.makeParamsValidator = exports.setParamType = exports.validateRequired = exports.paramTypeToString = exports.toString = exports.toBoolean = exports.toArray = exports.toNumber = exports.isNullOrUndefined = exports.printErrors = void 0;
const bind_di_1 = require("bind-di");
const controllermethodparams_1 = require("../consts/controllermethodparams");
const controllerparamtype_1 = __importDefault(require("../enums/controllerparamtype"));
const inputvalidationerror_1 = __importDefault(require("../errors/validation/inputvalidationerror"));
const dottedline_1 = require("../consts/dottedline");
const debug = require('debug')('bind:rest:runtime:validation');
const TAG = 'ParamsValidator';
const printErrors = (errors) => {
    return errors.map((e) => e.message).join('\n');
};
exports.printErrors = printErrors;
const isNullOrUndefined = (val) => val === undefined || val === null;
exports.isNullOrUndefined = isNullOrUndefined;
const toNumber = (i) => {
    if (exports.isNullOrUndefined(i))
        return undefined;
    const x = Number(i);
    if (Number.isNaN(x)) {
        return new TypeError();
    }
    return x;
};
exports.toNumber = toNumber;
const toArray = (i) => {
    if (exports.isNullOrUndefined(i))
        return undefined;
    if (!Array.isArray(i)) {
        return new TypeError();
    }
    return i;
};
exports.toArray = toArray;
const toBoolean = (s) => {
    if (exports.isNullOrUndefined(s))
        return undefined;
    if (typeof s === 'boolean') {
        return s;
    }
    if (typeof s !== 'string') {
        return new TypeError();
    }
    const lc = s.toLocaleLowerCase();
    if (lc === 'false' || lc === '0') {
        return false;
    }
    if (lc === 'true' || lc === '1') {
        return true;
    }
    return new TypeError();
};
exports.toBoolean = toBoolean;
const toString = (s) => {
    if (exports.isNullOrUndefined(s))
        return undefined;
    if (typeof s === 'string')
        return s;
    if (s instanceof String)
        return s.valueOf();
    if (typeof s === 'number')
        return String(s);
    return new TypeError();
};
exports.toString = toString;
const paramTypeToString = (paramType) => {
    let ret = exports.toString(paramType);
    if (paramType === null) {
        return 'Null';
    }
    if (paramType === undefined) {
        return 'undefined';
    }
    if (typeof ret === 'string') {
        return ret;
    }
    if (paramType.name) {
        return paramType.name;
    }
    try {
        ret = `${paramType}`;
        return ret;
    }
    catch (e) {
        return 'UNKNOWN_TYPE';
    }
};
exports.paramTypeToString = paramTypeToString;
function validateRequired(o) {
    const params = o.params.map((param, i) => {
        if (param instanceof Error)
            return param;
        if (o.meta[i] && o.meta[i].isRequired && exports.isNullOrUndefined(param)) {
            return new Error(`
      Required parameter not passed from request
      parameterType="${controllerparamtype_1.default[o.meta[i].paramDecoratorType]}" 
      parameterName="${o.meta[i].paramName}" 
      position="${i + 1}"`);
        }
        return param;
    });
    return { params, meta: o.meta };
}
exports.validateRequired = validateRequired;
function setParamType(o) {
    const params = o.params.map((param, i) => {
        var _a, _b;
        if (param instanceof Error)
            return param;
        let ret;
        if (!o.meta[i] || !o.meta[i].paramType) {
            debug('%s No meta or no paramType for param "%s"', TAG, (_a = o.meta[i]) === null || _a === void 0 ? void 0 : _a.paramName);
            ret = param;
        }
        else {
            switch (o.meta[i].paramType) {
                case controllermethodparams_1.PARAM_TYPE_STRING:
                    ret = exports.toString(param);
                    break;
                case controllermethodparams_1.PARAM_TYPE_NUMBER:
                    ret = exports.toNumber(param);
                    break;
                case controllermethodparams_1.PARAM_TYPE_BOOLEAN:
                    ret = exports.toBoolean(param);
                    break;
                case controllermethodparams_1.PARAM_TYPE_ARRAY:
                    ret = exports.toArray(param);
                    break;
                case controllermethodparams_1.PARAM_TYPE_OBJECT:
                    ret = param;
                    break;
                default:
                    if (o.meta[i].paramDecoratorType === controllerparamtype_1.default.Body) {
                        if (typeof param === 'string') {
                            debug('%s Cannot set prototype of body param to %s because body is string', TAG, exports.paramTypeToString(o.meta[i].paramType));
                            ret = new TypeError();
                        }
                        else {
                            debug('%s Setting prototype of Body param to "%s"', TAG, (_b = o.meta[i].paramType) === null || _b === void 0 ? void 0 : _b.name);
                            Reflect.setPrototypeOf(param, o.meta[i].paramType);
                            ret = param;
                        }
                    }
                    else if (param instanceof o.meta[i].paramType) {
                        ret = param;
                    }
                    else {
                        ret = new TypeError();
                    }
            }
        }
        if (ret instanceof TypeError) {
            return new Error(`
      Request parameter of type "${typeof param}" 
      cannot be converted to "${exports.paramTypeToString(o.meta[i].paramType)}"
      parameterType="${controllerparamtype_1.default[o.meta[i].paramDecoratorType]}" 
      parameterName="${o.meta[i].paramName}" 
      position="${i + 1}" 
      `);
        }
        return ret;
    });
    return { params, meta: o.meta };
}
exports.setParamType = setParamType;
function makeParamsValidator(meta, controllerName) {
    return function paramsValidator(params) {
        debug('%s entered paramsValidator for controller=%s with meta=%o params=%o', TAG, controllerName, meta, params);
        const res = [validateRequired, setParamType].reduce((acc, next) => {
            return next(acc);
        }, { params, meta });
        const errors = res.params.filter((param) => param instanceof Error);
        if (errors && errors.length > 0) {
            const message = exports.printErrors(errors);
            debug('%s have ValidationErrors %s', TAG, message);
            throw new inputvalidationerror_1.default(`Controller="${controllerName}"\nErrors:\n${message}`);
        }
        debug('%s paramsValidator returning params %o', TAG, res.params);
        return res.params;
    };
}
exports.makeParamsValidator = makeParamsValidator;
function makeValidateAsync(context, validators, controllerName) {
    const asyncValidators = validators.map((v) => v(context));
    return function validateParamsAsync(params) {
        debug('%s Entered validateParamsAsync with params=%o', TAG, params);
        const validationResults = params.map((param, i) => {
            if (asyncValidators[i]) {
                return asyncValidators[i](param);
            }
            return undefined;
        });
        return Promise.all(validationResults).then((results) => {
            debug('%s All validationResults are resolved %o', TAG, results);
            const errors = results.reduce((acc, next) => {
                if (bind_di_1.isDefined(next)) {
                    acc.push(next.message);
                }
                return acc;
            }, []);
            if (errors.length > 0) {
                debug('%s have asyncValidation errors %o', TAG, errors);
                throw new inputvalidationerror_1.default(`Controller=${controllerName}\n${errors.join(dottedline_1.DOTTED_LINE)}`);
            }
            else {
                debug('%s no async validation errors. params=%o', TAG, params);
                return params;
            }
        });
    };
}
exports.makeValidateAsync = makeValidateAsync;
//# sourceMappingURL=paramsvalidator.js.map