"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const bind_di_1 = require("bind-di");
const metaprops_1 = require("../decorators/metaprops");
const paramsvalidator_1 = require("./paramsvalidator");
const applicationerror_1 = __importDefault(require("../errors/applicationerror"));
const middlewares_1 = require("../decorators/middlewares");
const debug = require('debug')('bind:rest:runtime:controller');
const TAG = 'ControllerParser';
const joinPath = (base = '', callPath = '') => base + callPath;
function parseController(container) {
    return (component) => {
        var _a;
        const id = bind_di_1.stringifyIdentity(component.identity);
        if (!component.componentMetaData[metaprops_1.IS_CONTROLLER]) {
            debug('%s Component "%s" not a controller', TAG, id);
            return [];
        }
        if (!component.identity.clazz) {
            throw new Error(`Controller class not found for id="${id}"`);
        }
        const controllerPrototype = (_a = component.identity.clazz) === null || _a === void 0 ? void 0 : _a.prototype;
        if (!controllerPrototype) {
            throw new Error(`Controller class prototype not found for id="${id}"`);
        }
        const basePath = Reflect.getMetadata(metaprops_1.SYM_REQUEST_PATH, component.identity.clazz) || '';
        const constollerMiddlewares = Reflect.getMetadata(metaprops_1.SYM_CONTROLLER_MIDDLEWARES, component.identity.clazz) || [];
        debug('%s basePath for "%s" = "%s"', TAG, controllerPrototype.constructor.name, basePath);
        const props = Object.getOwnPropertyNames(controllerPrototype).filter((name) => name !== 'constructor');
        return props
            .map((p) => {
            var _a, _b;
            const controllerName = `${(_b = (_a = component.identity) === null || _a === void 0 ? void 0 : _a.clazz) === null || _b === void 0 ? void 0 : _b.name}.${p}`;
            let ctrlWithMiddleware;
            let controllerMiddleware;
            const metaMethods = Reflect.getMetadata(metaprops_1.SYM_REQUEST_METHOD, controllerPrototype, p);
            if (!bind_di_1.isDefined(metaMethods)) {
                debug('%s Method "%s" is NOT a controller. Returning null!!!', TAG, controllerName);
                return null;
            }
            const metaData = Reflect.getMetadata(bind_di_1.COMPONENT_META_DATA, controllerPrototype, p) || {};
            const paramsMeta = Reflect.getMetadata(metaprops_1.SYM_METHOD_PARAMS, controllerPrototype, p) || [];
            debug('%s paramsMeta=%o', TAG, paramsMeta);
            const metaPath = Reflect.getMetadata(metaprops_1.SYM_REQUEST_PATH, controllerPrototype, p) || '';
            let aMiddlewares = Reflect.getMetadata(metaprops_1.SYM_CONTROLLER_MIDDLEWARES, controllerPrototype, p) || [];
            aMiddlewares = constollerMiddlewares.concat(aMiddlewares);
            let controllerMiddlewareFactory;
            if (aMiddlewares.length > 0) {
                controllerMiddlewareFactory = middlewares_1.toMWFuncFactory(aMiddlewares);
                controllerMiddleware = controllerMiddlewareFactory(container);
            }
            const methods = Array.from(metaMethods);
            const paramExtractors = paramsMeta.map((meta) => meta.f(container));
            const validators = paramsMeta.map((meta) => {
                if (meta.validator) {
                    return meta.validator(container);
                }
                return (ctx) => (val) => undefined;
            });
            const validateParams = paramsvalidator_1.makeParamsValidator(paramsMeta, controllerName);
            const ctrl = (context) => {
                Reflect.set(context, 'controllerName', controllerName);
                const validateAsync = paramsvalidator_1.makeValidateAsync(context, validators, controllerName);
                const oCtrl = component.get([context]);
                debug('%s got oCtrl %o', TAG, oCtrl);
                let futureParams;
                if (paramExtractors.length > 0) {
                    debug('%s need futureParams', TAG);
                    futureParams = Promise.all(paramExtractors.map((f, i) => {
                        debug('%s calling paramExtractor func with context', TAG);
                        try {
                            return f(context);
                        }
                        catch (e) {
                            throw new applicationerror_1.default(`Error parsing controller parameters
                  Controller=${controllerName}
                  parameterIndex=${i}
                  Error=${e.message}`);
                        }
                    }))
                        .then((params) => {
                        debug('%s extracted futureParams=%o', TAG, params);
                        return params;
                    })
                        .catch((e) => {
                        debug('%s exception from futureParams %o', TAG, e);
                        throw new applicationerror_1.default(`Error Parsing request parameters
              Controller=${controllerName}
              Error=${e.message}`, e);
                    });
                }
                else {
                    debug('%s making empty futureParams', TAG);
                    futureParams = Promise.resolve([]);
                }
                return futureParams
                    .then(validateParams)
                    .then(validateAsync)
                    .then((validatedParams) => {
                    debug('%s got validatedParams %o', TAG, validatedParams);
                    Reflect.set(context, 'controllerArguments', validatedParams);
                    return oCtrl[p](...validatedParams);
                })
                    .catch((e) => {
                    debug('%s Error in Controller chain %o', TAG, e);
                    throw e;
                });
            };
            if (controllerMiddleware) {
                debug('%s adding controller middleware for controller="%s"', TAG, controllerName);
                ctrlWithMiddleware = (context) => {
                    return controllerMiddleware(context).then(ctrl);
                };
            }
            const priority = bind_di_1.isDefined(metaData[metaprops_1.CONTROLLER_MATCHER]) ? 1 : 0;
            return {
                name: controllerName,
                requestMethods: methods,
                routePath: joinPath(basePath, metaPath),
                ctrl: ctrlWithMiddleware || ctrl,
                matcher: metaData[metaprops_1.CONTROLLER_MATCHER],
                priority,
            };
        })
            .filter(bind_di_1.notEmpty);
    };
}
exports.default = parseController;
//# sourceMappingURL=parsecontroller.js.map