"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.makeRequest = exports.setServiceOptions = void 0;
const http_1 = __importDefault(require("http"));
const https_1 = __importDefault(require("https"));
const timers_1 = require("timers");
const util_1 = __importDefault(require("util"));
const v4_1 = __importDefault(require("uuid/v4"));
const querystring_1 = __importDefault(require("querystring"));
const is_stream_1 = __importDefault(require("is-stream"));
const tunnel_1 = require("./tunnel");
const applicationerror_1 = __importDefault(require("../errors/applicationerror"));
const IncomingMessageResponse_1 = __importDefault(require("../httpresponse/IncomingMessageResponse"));
const reponsetimeout_1 = __importDefault(require("../errors/http/reponsetimeout"));
const debug = require('debug')('bind:rest:request');
const REQUEST_ID = 'requestID';
const STATUS_CODE = 'statusCode';
const REQUEST_HOST = 'requestHost';
const REQUEST_URI = 'requestURI';
const TIMEOUT_VALUE = 'timeoutValue';
const REQUEST_ERROR = 'RequestError';
const REQUEST_METHOD = 'requestMethod';
const ELAPSED_TIME = 'elapsedTime';
const RESPONSE_HEADERS = 'responseHeaders';
const DEFAULT_TIMEOUT = 5000;
const TAG = 'makeRequest';
function setServiceOptions(options) {
    debug(TAG, 'Entered setServiceOptions with options: ', options);
    if (!options.requestID) {
        options.requestID = v4_1.default();
    }
    if (!options.requestOptions.method) {
        debug(TAG, "Missing requestOptions.method  Setting default method 'GET'");
        options.requestOptions.method = 'GET';
    }
    if (options.basicAuth) {
        if (!options.requestOptions.headers) {
            options.requestOptions.headers = {};
        }
        options.requestOptions.headers.Authorization = `Basic ${Buffer.from(`${options.basicAuth.user}:${options.basicAuth.password}`, 'utf8').toString('base64')}`;
        debug('%s Added AuthorizationHeader=%s', TAG, options.requestOptions.headers.Authorization);
    }
    if (options.RequestType === 'HTTPS' ||
        (options.requestOptions.protocol &&
            options.requestOptions.protocol.toLocaleLowerCase() === 'https:')) {
        options.requestOptions.protocol = 'https:';
        debug(TAG, 'INSIDE HTTPS setServiceOptions with options: ', options);
        if (!options.requestOptions.port) {
            options.requestOptions.port = 443;
        }
        if (options.disableCertificateValidation === true) {
            options.requestOptions.rejectUnauthorized = false;
            debug(TAG, 'added rejectUnauthorized');
        }
        if (options.proxy) {
            debug('Will setup agent with https over proxy details');
            options.requestOptions.agent = tunnel_1.getHTTPSOverHTTPTunnel(options);
        }
        else {
            debug(TAG, 'No PROXY in options. Setting up default agent');
            options.requestOptions.agent = new https_1.default.Agent(options.requestOptions);
        }
    }
    else {
        debug(TAG, 'Not HTTPS. Setting options.requestOptions.protocol to http');
        options.requestOptions.protocol = 'http:';
        if (options.proxy) {
            options.requestOptions.agent = tunnel_1.getHTTPOverHTTPTunnel(options);
        }
        else {
            options.requestOptions.agent = false;
        }
    }
    if (options.requestOptions.port) {
        options.requestOptions.port = Number(options.requestOptions.port);
        if (Number.isNaN(options.requestOptions.port)) {
            throw new TypeError(`${TAG} value of options.requestOptions.port could not be converted to Number. Value was="${options.requestOptions.port}"`);
        }
        debug(TAG, 'Port number set to=', options.requestOptions.port);
    }
    if (!options.requestOptions.hostname && options.requestOptions.host) {
        options.requestOptions.hostname = options.requestOptions.host;
    }
    if (options.requestOptions.query) {
        const queryString = querystring_1.default.stringify(options.requestOptions.query);
        if (options.requestOptions.path.endsWith('?')) {
            options.requestOptions.path = options.requestOptions.path.slice(0, -1);
        }
        options.requestOptions.path = `${options.requestOptions.path}?${queryString}`;
    }
    return options;
}
exports.setServiceOptions = setServiceOptions;
function makeRequest(requestOptions) {
    let ClientRequestorObject;
    let timeout = DEFAULT_TIMEOUT;
    let options;
    try {
        options = setServiceOptions(requestOptions);
    }
    catch (e) {
        if (requestOptions.logger) {
            requestOptions.logger.error(TAG, 'Failed to setServiceOptions', e);
        }
        return Promise.reject(e);
    }
    debug('^^^^^ options after setServiceOptions: ', options);
    if (options.requestOptions.protocol === 'https:') {
        ClientRequestorObject = https_1.default;
        debug('@@@@@@@@@@@ USING HTTPS @@@@@@@@@@@@@');
    }
    else {
        ClientRequestorObject = http_1.default;
        debug('@@@@@@@@@@@ USING HTTP @@@@@@@@@@@@@');
    }
    let startTime;
    if (options.timeout) {
        timeout = Number(options.timeout);
        if (Number.isNaN(timeout)) {
            if (options.logger) {
                options.logger.error(TAG, `Value of options.timeout=${options.timeout} cannot be converted to number. Using default falue`);
            }
            timeout = DEFAULT_TIMEOUT;
        }
    }
    const myport = options.requestOptions.port ? `:${options.requestOptions.port}` : '';
    const myuri = `${options.requestOptions.protocol}//${options.requestOptions.hostname}${myport}${options.requestOptions.path}`;
    return new Promise((resolve, reject) => {
        let request;
        let resolved;
        let timeoutOccured = false;
        if (typeof options.payload === 'string') {
            const payloadLen = Buffer.byteLength(options.payload);
            debug('%s set payloadLen=%s', TAG, payloadLen);
            if (!options.requestOptions.headers) {
                options.requestOptions.headers = [];
            }
            options.requestOptions.headers['Content-Length'] = payloadLen;
            if (options.logger) {
                options.logger.debug(TAG, 'payload is string. Method=', options.requestOptions.method, 'headers=', options.requestOptions.headers);
            }
        }
        const timeoutId = setTimeout(() => {
            timeoutOccured = true;
            try {
                if (request) {
                    request.abort();
                }
            }
            catch (e) {
                debug(TAG, ' Failed to abort request: %o', e);
                if (options.logger) {
                    options.logger.error(`${TAG}=Failed to abort request. ${REQUEST_ID}=${options.requestID} ${REQUEST_ERROR}=`, e);
                }
            }
        }, timeout);
        try {
            if (options.logger) {
                options.logger.info(`${TAG}=Start ${TIMEOUT_VALUE}=${timeout} milliseconds. ${REQUEST_ID}=${options.requestID} ${REQUEST_HOST}=${options.requestOptions.hostname} ${REQUEST_URI}="${myuri}" ${REQUEST_METHOD}=${options.requestOptions.method}`);
                options.logger.debug(`${TAG}=Start ${REQUEST_ID}=${options.requestID} RequestHeaders:`, options.requestOptions.headers);
            }
            startTime = Date.now();
            request = ClientRequestorObject.request(options.requestOptions);
            debug(TAG, ' Created request object');
            request.on('error', (e) => {
                try {
                    timers_1.clearTimeout(timeoutId);
                }
                catch (ex) {
                    if (options.logger) {
                        options.logger.error(`${TAG}=clearTimeoutFailed on error for ${REQUEST_ID}=${options.requestID} ${REQUEST_HOST}=${options.requestOptions.hostname} ${REQUEST_URI}="${myuri}" ${REQUEST_METHOD}=${options.requestOptions.method} error=${e.message} ${TIMEOUT_VALUE}=${timeout} milliseconds`);
                    }
                }
                if (!timeoutOccured) {
                    if (options.logger) {
                        options.logger.error(`${TAG}=ErrorEvent ${REQUEST_ID}=${options.requestID}  ${REQUEST_HOST}=${options.requestOptions.hostname}" ${REQUEST_URI}="${myuri}" ${REQUEST_METHOD}="${options.requestOptions.method}" error=`, e);
                    }
                    reject(new Error(`There is an error in the get request ${util_1.default.inspect(e)}`));
                }
                else {
                    debug(TAG, ' HTTP REQUEST ERROR ', e.message);
                    if (options.logger) {
                        options.logger.error(`${TAG}=ErrorEvent with timeout. ${REQUEST_ID}=${options.requestID}  ${REQUEST_HOST}=${options.requestOptions.hostname} ${REQUEST_URI}="${myuri}" ${REQUEST_METHOD}="${options.requestOptions.method}" error=`, e);
                    }
                    if (!resolved) {
                        reject(new Error(`HTTP Request error ${util_1.default.inspect(e)}`));
                    }
                    else {
                        debug(TAG, ' Promise was already resolved');
                        if (options.logger) {
                            options.logger.debug(TAG, `error but already resolved. ${REQUEST_ID}=${options.requestID}  ${REQUEST_HOST}=${options.requestOptions.hostname} ${REQUEST_URI}="${myuri}" ${REQUEST_METHOD}=${options.requestOptions.method}`);
                        }
                    }
                }
            });
            request.on('abort', () => {
                try {
                    timers_1.clearTimeout(timeoutId);
                }
                catch (e) {
                    if (options.logger) {
                        options.logger.error(`${TAG}=clearTimeoutFailed on abort event for ${REQUEST_ID}=${options.requestID} ${REQUEST_HOST}=${options.requestOptions.hostname} ${REQUEST_URI}="${myuri}" ${REQUEST_METHOD}="${options.requestOptions.method}" error=${e.message} ${TIMEOUT_VALUE}=${timeout} milliseconds`);
                    }
                }
                debug(TAG, 'ON-ABORT CALLED');
                if (options.logger) {
                    options.logger.debug(TAG, 'request aborted for requestID=', options.requestID, ' REQUEST_HOST=', options.requestOptions.hostname, ' requestURI="', myuri, ' REQUEST_METHOD=', options.requestOptions.method, ' TIMEOUT_VALUE=', timeout, ' milliseconds');
                }
                if (timeoutOccured) {
                    if (options.logger) {
                        options.logger.debug(`${TAG}=debug resolving timed-out request. ${REQUEST_ID}=${options.requestID} ${REQUEST_HOST}=${options.requestOptions.hostname} ${REQUEST_URI}="${myuri}" ${REQUEST_METHOD}=${options.requestOptions.method} ${TIMEOUT_VALUE}=${timeout} milliseconds`);
                    }
                    if (!resolved) {
                        resolved = true;
                        reject(new reponsetimeout_1.default(`The request has taken longer than the allotted ${options.timeout} milliseconds`, options));
                    }
                }
                else {
                    reject(new Error(`Http Request aborted`));
                }
            });
            request.on('response', (response) => {
                try {
                    timers_1.clearTimeout(timeoutId);
                }
                catch (ex) {
                    if (options.logger) {
                        options.logger.error(`${TAG}=clearTimeoutFailed for ${REQUEST_ID}=${options.requestID} ${REQUEST_HOST}=${options.requestOptions.hostname} ${REQUEST_URI}="${myuri}" ${REQUEST_METHOD}=${options.requestOptions.method} ${REQUEST_ERROR}=${ex.message} ${TIMEOUT_VALUE}=${timeout} milliseconds`);
                    }
                }
                if (options.logger) {
                    options.logger.debug(`${TAG}=onresponse ${REQUEST_ID}=${options.requestID} ${STATUS_CODE}=${response.statusCode} ${REQUEST_HOST}=${options.requestOptions.hostname} ${REQUEST_URI}="${myuri}" ${REQUEST_METHOD}=${options.requestOptions.method}`);
                }
                debug('%s onResponse called with statusCode=%s', TAG, response.statusCode);
                if (!resolved) {
                    resolved = true;
                    resolve(new IncomingMessageResponse_1.default(response, options.requestID));
                }
            });
            request.on('end', (response) => {
                debug('ON-END CALLED with response="%s"', !!response);
                if (options.logger) {
                    options.logger.info(`${TAG}=end ${REQUEST_ID}=${options.requestID} ${REQUEST_HOST}=${options.requestOptions.hostname} ${REQUEST_URI}="${myuri}" ${REQUEST_METHOD}=${options.requestOptions.method}`);
                }
            });
            request.on('upgrade', (response) => {
                debug('ON-UPGRADE CALLED with resonse="%s"', !!response);
                if (options.logger) {
                    options.logger.info(`${TAG}=upgrade ${REQUEST_ID}=${options.requestID} ${REQUEST_HOST}=${options.requestOptions.hostname} ${REQUEST_URI}="${myuri}" ${REQUEST_METHOD}=${options.requestOptions.method}`);
                }
            });
            let payloadPiped = false;
            if (options.requestOptions.method === 'POST' || options.requestOptions.method === 'PUT') {
                if (options.payload) {
                    debug('%s Request method=%s has payload=%s', TAG, options.requestOptions.method);
                    if (typeof options.payload === 'string') {
                        debug(TAG, 'Request payload is a string: %s', options.payload);
                        if (options.logger) {
                            options.logger.debug(TAG, 'Request payload is a string. Method=', options.requestOptions.method, options.requestOptions.headers);
                        }
                        request.write(options.payload);
                        debug(TAG, 'Written payload to request');
                    }
                    else {
                        if (options.logger) {
                            options.logger.debug(TAG, 'Request payload is a stream=.', is_stream_1.default.readable(options.payload), ' method=', options.requestOptions.method, options.requestOptions.headers);
                        }
                        debug(TAG, 'Request payload is a stream');
                        options.payload.pipe(request);
                        payloadPiped = true;
                    }
                }
                else {
                    debug(TAG, 'NO payload in requestOptions');
                }
            }
            if (!payloadPiped) {
                debug(TAG, 'Calling request.end()');
                request.end();
            }
        }
        catch (ex) {
            if (options.logger) {
                options.logger.error(`${TAG}=Failed to create request object ${REQUEST_ID}=${options.requestID} ${REQUEST_ERROR}="${ex.message}"`);
            }
            reject(new applicationerror_1.default(`${TAG} request object ${REQUEST_ERROR}="${ex.message}"`));
        }
    })
        .then((resp) => {
        const endTime = Date.now();
        const elapsedTime = endTime - startTime;
        debug('%s Request completed for %s=%s %s=%s %s=%s', TAG, REQUEST_ID, options.requestID, ELAPSED_TIME, elapsedTime, STATUS_CODE, resp.statusCode);
        if (options.logger) {
            options.logger.info(`${TAG}=OK Request complete. ${REQUEST_ID}=${options.requestID} ${REQUEST_HOST}=${options.requestOptions.hostname} ${REQUEST_URI}="${myuri}" ${REQUEST_METHOD}=${options.requestOptions.method} ${STATUS_CODE}=${resp.statusCode} ${RESPONSE_HEADERS}=`, resp.headers, `${ELAPSED_TIME}=${elapsedTime} milliseconds`);
        }
        return resp;
    })
        .catch((e) => {
        const endTime = Date.now();
        const elapsedTime = endTime - startTime;
        debug('%s Request FAILED for %s=%s %s=%s %s=%o', TAG, REQUEST_ID, options.requestID, ELAPSED_TIME, elapsedTime, REQUEST_ERROR, e);
        if (options.logger) {
            options.logger.error(`${TAG}=Exception  ${REQUEST_ID}=${options.requestID} ${REQUEST_HOST}=${options.requestOptions.hostname} ${REQUEST_URI}="${myuri}" ${REQUEST_METHOD}=${options.requestOptions.method} Error="${e.message}" ${ELAPSED_TIME}=${elapsedTime} milliseconds. Exception=`, e);
        }
        throw e;
    });
}
exports.makeRequest = makeRequest;
//# sourceMappingURL=request.js.map